function histMatchFeatR(outFile,inFile,alpha,resynth,delta_dd)
%histMatchFeatR(outFile,inFile,alpha,resynth,delta_dd)
%   inFile should be a 16kHz mono 16-bit uncompressed (PCM-format) MS-WAV 
%   file (with extension .wav), or a 16kHz mono 16-bit uncompressed SPHERE 
%   file (with any extension other than .wav, e.g. .sph, .wv1, .wvs, and so 
%   on).
%
%   The input file header (1024 bytes for SPHERE, 44 bytes for MS-WAV) is 
%   skipped automatically based on the file extension. If any decompression 
%   is needed (Shorten, etc.) the program must be modified.
%
%   outFile is written as a SRI-format ASCII file, in the form directly
%   convertible to the SRI binary using feacat. E.g., as
%
%   0   0   C1 C2 C3 C4 ... C39
%   0   1   C1 C2 C3 C4 ... C39
%   0   2   C1 C2 C3 C4 ... C39
%            ...
%   0  N-1  C1 C2 C3 C4 ... C39
%
%   for a file with N time frames.
%
%   The value of alpha is a program parameter that controls the intensity of
%   the noise compensation process. alpha must be between 0 and 1, inclusive. 
%   If alpha is 0, the noise compensation is very intense, and this typically 
%   significantly degrades ASR accuracy and intelligibility. If the value of
%   alpha is 1, then no noise compensation is performed. The standard value of 
%   alpha, which currently works best on CMU Sphinx, is 0.85. The best value of 
%   alpha may vary across different speech recognizers (and/or databases). A 
%   more robust way of automatically selecting an optimal value of alpha in a 
%   general case is under development. 
%
%   This program can compute delta and double-delta coefficients. For delta, 
%   we compute d[n] = c[n+2] - c[n-2]. For double-delta, we  compute dd[n] = 
%   d[n+1] - d[n-1]. This follows the convention in "Spoken Language Processing" 
%   by Huang, Acero & Hon, 2001. CMN is performed before the delta calculations 
%   and is not applied to the delta and double-delta coefficients.
%
%   The delta and double-delta calculation can be skipped by setting the last 
%   input parameter equal to 0. In this case, each feature vector has 13 elements, 
%   as opposed to 39.
%
%   If desired, outFile can be written as a mono 16-bit 16kHz MS-WAV file, 
%   instead of a feature. This is useful to impose further processing on 
%   the signal before feature extraction, or to extract a different type of
%   feature than the one automatically generated in this script. See below
%   for more details.
%   
%   To resynthesize speech, please set resynth equal to 1. If resynth is 
%   equal to anything else, feature extraction is performed as 
%   described below (and resynthesis is performed along the way, without 
%   writing the audio). If resynth is equal to 1, the value of delta_dd
%   (the last input parameter) has no effect, but still must be specified.
%
%   For correct execution, ALL input parameters must be specified. The standard
%   values are as follows:
%           outFile  -> output filename
%           inFile   -> input filename
%           alpha    -> 0.85
%           resynth  -> 0 (or 1 for resynthesis and no feature)
%           delta_dd -> 1 (or 0 for no deltas & double-deltas)
%
%   The feature extraction included in this script is called SPNCC
%   (Simple Power-Normalized Cepstral Coefficients). It is a simpler
%   variant of PNCC, and varies in only two distinct ways from MFCC.
%
%   Simple Power-Normalized Cepstral Coefficients:
%   
%   input audio --> [pre-emphasis] --> [|STFT|.^2] --> 
%   [Gammatone filter bank] --> [peak power noramlization] --> 
%   [power-law nonlinearity, P.^(1/15)] --> [DCT] --> [CMN] --> feature
%
%   W.r.t. MFCC, the conventional triangular filter bank is replaced by
%   Slaney's Gammatone filter bank, and the logarithmic nonlinearity is
%   replaced by a power-law (or, root-law) with exponent 1/15 (i.e. x.^(1/15)).
%
%   W.r.t. PNCC, no medium-time processing (i.e. for noise reduction) is 
%   performed.
%
%   Contact: 
%   mharvill@cs.cmu.edu (Mark Harvilla) 
%   rms@cs.cmu.edu (Prof. Richard Stern)
%
%   Written by Mark Harvilla
%   Carnegie Mellon University
%   August 5, 2011

%%%%%%%%%%%%
% program
% parameters
fs        = 16000;
pow_expon = 1/15;
T_wind    = 0.0512;
T_shift   = 0.0128;
iFFTSize  = 1024;

if isdeployed
   alpha    = str2double(alpha);
   resynth  = str2double(resynth);
   delta_dd = str2double(delta_dd);
end

%%%%%%%%%%%%
% load pre-computed power signal histograms
% and Gammatone filter bank
load avgHistMaxMin_SPNCC_NoCMN_sri-wsj.mat;
load GammatoneFilterBank.mat;

%%%%%%%%%%%%
% read input
% waveform
fid = fopen(inFile, 'rb');
if strcmp(inFile(end-2:end),'wav') == 1 % <-- MS-WAV format
    fseek(fid, 44, 'bof');
else % <-- else assumed to be SPHERE format
    fseek(fid, 1024, 'bof'); 
end
ad_x  = fread(fid, 'int16');
fclose(fid);
ad_x = ad_x/32768;

%%%%%%%%%%%%
% downsample to 
% 8kHz
%ad_x = resample(ad_x,1,2);

%%%%%%%%%%%%
% conventional
% pre-emphasis
ad_x = filter([1 -0.97], 1, ad_x);

%%%%%%%%%%%%
% STFT
%
N1 = floor(T_wind*fs);
N2 = floor(T_shift*fs);
W  = hamming(N1);

iNumWindows = 1 + floor((length(ad_x) - N1)/N2);
X           = zeros((iFFTSize/2) + 1, iNumWindows);
Xp          = X;

for i_FI=1:iNumWindows
    x      = ad_x( 1 + (i_FI-1)*N2 : N1 + (i_FI-1)*N2 ).* W;
    Xframe = fft(x,iFFTSize); 
    Xframe = Xframe(1:(iFFTSize/2) + 1);
    
    Xp(:,i_FI) = Xframe;
    X(:,i_FI)  = abs(Xframe);
end

%%%%%%%%%%%%
% compute filter 
% bank output
P = (transpose(H).^2)*(X.^2);

%%%%%%%%%%%%
% peak power 
% normalization
adSumPower = sum(P,2);
adSorted   = sort(adSumPower);
dMaxPower  = adSorted(round(0.98 * length(adSumPower)));
U          = (P / dMaxPower) * 1e10;

%%%%%%%%%%%%
% power-law
% nonlinearity
Pnorm = U.^pow_expon;

%%%%%%%%%%%%
% histogram-based
% processing
Pproc = histMfeat(Pnorm,pow_expon,alpha,hgram);

%%%%%%%%%%%%
% power spectrum
% weights
w = Pproc./P;
%w = sqrt(Pproc./P);

%%%%%%%%%%%%
% weight
% smoothing
[iMax,jMax] = size(w);
wS          = zeros(iMax,jMax);
M = 1; %frequency smoothing (half - 1/2) range
N = 1; %time smoothing (half - 1/2) range
for i=1:iMax
    for j=1:jMax
            wS(i,j) = mean(mean(w( max(1,i-M):min(i+M,iMax),...
                max(1,j-N):min(j+N,jMax) )));
    end
end

%%%%%%%%%%%%
% spectrogram
% reshaping
Hs = sum(H,2); Hs(Hs==0)=eps;
u  = H*wS;
u  = u./repmat(Hs,1,size(u,2));
Y  = u.*Xp;

%%%%%%%%%%%%
% reconstruct
% using OLA
y   = zeros(size(ad_x));
x_t = ifft([Y;conj(flipud(Y(2:end-1,:)))]); %to time-domain
for i_FI = 1:iNumWindows
    x = x_t(1:N1,i_FI);
    y(1 + (i_FI-1)*N2 : N1 + (i_FI-1)*N2) =...
        y(1 + (i_FI-1)*N2 : N1 + (i_FI-1)*N2) + (x .* W);
end

%%%%%%%%%%%%
% post
% de-emphasis
%y = filter(1, [1 -0.97], y);

%%%%%%%%%%%%
% upsample
% to 16kHz
%y = resample(y,2,1);

%%%%%%%%%%%%
% feature extraction OR
% resynthesis
y = 0.9*(y / max(abs(y)));
if resynth == 1
    wavwrite(y,fs,16,outFile);
else
    %%%%%%%%%%%% <------- feature extraction (SPNCC) -------> %%%%%%%%%%%%
    ad_x = y;
    clearvars -except ad_x H outFile x_orig delta_dd;
    %%%%%%%%%%%%
    % program
    % parameters
    fs        = 16000;
    pow_expon = 1/15;
    T_wind    = 0.0256;
    T_shift   = 0.01;
    iFFTSize  = 1024;

    %%%%%%%%%%%%
    % conventional
    % pre-emphasis
    ad_x = filter([1 -0.97], 1, ad_x);

    %%%%%%%%%%%%
    % STFT
    %
    N1 = floor(T_wind*fs);
    N2 = floor(T_shift*fs);
    W  = hamming(N1);

    iNumWindows = 1 + floor((length(ad_x) - N1)/N2);
    X           = zeros((iFFTSize/2) + 1, iNumWindows);
    Xp          = X;

    for i_FI=1:iNumWindows
        x      = ad_x( 1 + (i_FI-1)*N2 : N1 + (i_FI-1)*N2 ).* W;
        Xframe = fft(x,iFFTSize); 
        Xframe = Xframe(1:(iFFTSize/2) + 1);

        Xp(:,i_FI) = Xframe;
        X(:,i_FI)  = abs(Xframe);
    end

    %%%%%%%%%%%%
    % compute filter 
    % bank output
    P = (transpose(H).^2)*(X.^2);

    %%%%%%%%%%%%
    % peak power 
    % normalization
    adSumPower = sum(P,2);
    adSorted   = sort(adSumPower);
    dMaxPower  = adSorted(round(0.98 * length(adSumPower)));
    U          = (P / dMaxPower) * 1e10;

    %%%%%%%%%%%%
    % power-law
    % nonlinearity
    Pnorm = U.^pow_expon;

    %%%%%%%%%%%%
    % DCT
    %
    C  = dct(Pnorm);

    %%%%%%%%%%%%
    % CMN, delta &
    % double-delta computation
    C  = C(1:13,:);
    C  = C - repmat(mean(C,2),1,size(C,2));
    if delta_dd ~= 0
       Cd = [C(:,3:end),zeros(size(C,1),2)] -...
           [zeros(size(C,1),2),C(:,1:end-2)];
       Cdd = [Cd(:,2:end),zeros(size(Cd,1),1)] -...
           [zeros(size(Cd,1),1),Cd(:,1:end-1)];
       %%%%%%%%%%%%
       % write the 
       % feature (ASCII format)
       Ct = [C;Cd;Cdd];
    else
       Ct = C;
    end
    fid = fopen(outFile, 'wb', 'ieee-be');
    for i = 1:size(Ct,2);
         for j = 1:size(Ct,1)
            if j == size(Ct,1) && i == size(Ct,2)
                fprintf(fid,'%.6f',Ct(j,i));
            elseif j == size(Ct,1)
                fprintf(fid,'%.6f\n',Ct(j,i));
            elseif j == 1
                fprintf(fid,'0 %d %.6f ',i-1,Ct(j,i));
            else
                fprintf(fid,'%.6f ',Ct(j,i));
            end
         end
    end
    fclose(fid);
end
